"""
parsing.py

Parses X12 837 005010X223A3 segments into a transactional domain model.

The parsing module includes a specific parser for the 837 transaction and loop parsing functions to create new loops
as segments are streamed to the transactional data model.

Loop parsing functions are implemented as set_[description]_loop(context: X12ParserContext, segment_data: Dict).
"""

from enum import Enum
from x12.parsing import match, X12ParserContext
from typing import Dict


class TransactionLoops(str, Enum):
    """
    The loops used to support the 837 005010X223A3 format.
    """

    HEADER = "header"
    SUBMITTER_NAME = "loop_1000a"
    RECEIVER_NAME = "loop_1000b"
    BILLING_PROVIDER = "loop_2000a"
    BILLING_PROVIDER_NAME = "loop_2010aa"
    BILLING_PROVIDER_PAY_TO_ADDRESS = "loop_2010ab"
    BILLING_PROVIDER_PAY_TO_PLAN_NAME = "loop_2010ac"
    SUBSCRIBER = "loop_2000b"
    SUBSCRIBER_NAME = "loop_2010ba"
    SUBSCRIBER_PAYER_NAME = "loop_2010bb"
    PATIENT = "loop_2000c"
    PATIENT_NAME = "loop_2010ca"
    CLAIM_INFORMATION = "loop_2300"
    CLAIM_ATTENDING_PROVIDER_NAME = "loop_2310a"
    CLAIM_OPERATING_PHYSICIAN_NAME = "loop_2310b"
    CLAIM_OTHER_OPERATING_PHYSICIAN_NAME = "loop_2310c"
    CLAIM_RENDERING_PROVIDER_NAME = "loop_2310d"
    CLAIM_SERVICE_FACILITY_LOCATION_NAME = "loop_2310e"
    CLAIM_REFERRING_PROVIDER_NAME = "loop_2310f"
    CLAIM_OTHER_SUBSCRIBER_INFORMATION = "loop_2320"
    CLAIM_OTHER_SUBSCRIBER_NAME = "loop_2330a"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_NAME = "loop_2330b"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_ATTENDING_PROVIDER_NAME = "loop_2330c"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_OPERATING_PHYSICIAN_NAME = "loop_2330d"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_OTHER_OPERATING_PHYSICIAN = "loop_2330e"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_SERVICE_FACILITY_LOCATION = "loop_2330f"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_RENDERING_PROVIDER_NAME = "loop_2330g"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_REFERRING_PROVIDER_NAME = "loop_2330h"
    CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_BILLING_PROVIDER_NAME = "loop_2330i"
    CLAIM_SERVICE_LINE = "loop_2400"
    CLAIM_SERVICE_LINE_DRUG_IDENTIFICATION_NAME = "loop_2410"
    CLAIM_SERVICE_LINE_OPERATING_PHYSICIAN_NAME = "loop_2420a"
    CLAIM_SERVICE_LINE_OTHER_OPERATING_PHYSICIAN_NAME = "loop_2420b"
    CLAIM_SERVICE_LINE_RENDERING_PROVIDER_NAME = "loop_2420c"
    CLAIM_SERVICE_LINE_REFERRING_PROVIDER_NAME = "loop_2420d"
    CLAIM_SERVICE_LINE_LINE_ADJUDICATION_INFORMATION = "loop_2430"
    FOOTER = "footer"


def _get_header(context: X12ParserContext) -> Dict:
    """Returns the 837 transaction header"""
    return context.transaction_data[TransactionLoops.HEADER]


def _get_billing_provider(context: X12ParserContext) -> Dict:
    """Returns the current billing provider record"""
    return context.transaction_data[TransactionLoops.BILLING_PROVIDER][-1]


def _is_patient_a_dependent(patient_record: Dict):
    """Returns true if the patient is a dependent record"""
    return patient_record.get("hl_segment", {}).get("hierarchical_level_code") == "23"


def _get_claim(context: X12ParserContext) -> Dict:
    """Returns the current claim record for the patient (either subscriber or dependent)"""
    if _is_patient_a_dependent(context.patient_record):
        claim = context.patient_record[TransactionLoops.CLAIM_INFORMATION][-1]
    else:
        claim = context.subscriber_record[TransactionLoops.CLAIM_INFORMATION][-1]
    return claim


def _get_other_subscriber(context: X12ParserContext) -> Dict:
    """Returns the other subscriber record for the current healthcare claim"""
    claim = _get_claim(context)
    return claim.get(TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION, [{}])[-1]


def _get_service_line(context: X12ParserContext) -> Dict:
    """Returns the current claim service line for the patieht (either subscriber or dependent)"""
    claim = _get_claim(context)
    return claim[TransactionLoops.CLAIM_SERVICE_LINE][-1]


@match("ST")
def set_header_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set header loop for the 837 transaction set.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """

    context.set_loop_context(
        TransactionLoops.HEADER, context.transaction_data[TransactionLoops.HEADER]
    )


@match("NM1", conditions={"entity_identifier_code": "41"})
def set_submitter_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Set the Header Submitter Name Loop 1000A

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.HEADER:
        context.transaction_data[TransactionLoops.SUBMITTER_NAME] = {"per_segment": []}
        submitter_loop = context.transaction_data[TransactionLoops.SUBMITTER_NAME]
        context.set_loop_context(TransactionLoops.SUBMITTER_NAME, submitter_loop)


@match("NM1", conditions={"entity_identifier_code": "40"})
def set_receiver_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Set the Header Receiver Name Loop 1000B

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.SUBMITTER_NAME:
        context.transaction_data[TransactionLoops.RECEIVER_NAME] = {}
        receiver_loop = context.transaction_data[TransactionLoops.RECEIVER_NAME]
        context.set_loop_context(TransactionLoops.RECEIVER_NAME, receiver_loop)


@match("HL", conditions={"hierarchical_level_code": "20"})
def set_billing_provider_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the Billing Provider Loop 2000A

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if TransactionLoops.BILLING_PROVIDER not in context.transaction_data:
        context.transaction_data[TransactionLoops.BILLING_PROVIDER] = []

    context.transaction_data[TransactionLoops.BILLING_PROVIDER].append({})
    billing_provider_loop = context.transaction_data[TransactionLoops.BILLING_PROVIDER][
        -1
    ]
    context.set_loop_context(TransactionLoops.BILLING_PROVIDER, billing_provider_loop)


@match("NM1", conditions={"entity_identifier_code": "85"})
def set_billing_provider_name_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the Billing Provider Name Loop 2010AA

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.BILLING_PROVIDER:
        billing_provider = _get_billing_provider(context)

        if TransactionLoops.BILLING_PROVIDER_NAME not in billing_provider:
            billing_provider[TransactionLoops.BILLING_PROVIDER_NAME] = {
                "per_segment": [],
            }
        billing_provider_name_loop = billing_provider[
            TransactionLoops.BILLING_PROVIDER_NAME
        ]
        context.set_loop_context(
            TransactionLoops.BILLING_PROVIDER_NAME, billing_provider_name_loop
        )


@match("NM1", conditions={"entity_identifier_code": "87"})
def set_pay_to_address_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the Billing Provider Pay to Address Name Loop 2010AB

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if "loop_2010a" in context.loop_name:
        billing_provider = _get_billing_provider(context)

        if TransactionLoops.BILLING_PROVIDER_PAY_TO_ADDRESS not in billing_provider:
            billing_provider[TransactionLoops.BILLING_PROVIDER_PAY_TO_ADDRESS] = {}
            pay_to_address_loop = billing_provider[
                TransactionLoops.BILLING_PROVIDER_PAY_TO_ADDRESS
            ]
            context.set_loop_context(
                TransactionLoops.BILLING_PROVIDER_PAY_TO_ADDRESS, pay_to_address_loop
            )


@match("NM1", conditions={"entity_identifier_code": "PE"})
def set_pay_to_plan_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the Billing Provider Pay to Plan Name Loop 2010AC

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if "loop_2010a" in context.loop_name:
        billing_provider = _get_billing_provider(context)

        if TransactionLoops.BILLING_PROVIDER_PAY_TO_PLAN_NAME not in billing_provider:
            billing_provider[TransactionLoops.BILLING_PROVIDER_PAY_TO_PLAN_NAME] = {
                "ref_segment": [],
            }
            pay_to_plan_loop = billing_provider[
                TransactionLoops.BILLING_PROVIDER_PAY_TO_PLAN_NAME
            ]
            context.set_loop_context(
                TransactionLoops.BILLING_PROVIDER_PAY_TO_PLAN_NAME, pay_to_plan_loop
            )


@match("HL", conditions={"hierarchical_level_code": "22"})
def set_subscriber_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the subscriber loop 2000B

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """

    billing_provider = _get_billing_provider(context)

    if TransactionLoops.SUBSCRIBER not in billing_provider:
        billing_provider[TransactionLoops.SUBSCRIBER] = []

    billing_provider[TransactionLoops.SUBSCRIBER].append({})
    context.subscriber_record = billing_provider[TransactionLoops.SUBSCRIBER][-1]
    context.set_loop_context(TransactionLoops.SUBSCRIBER, context.subscriber_record)

    if context.hl_segment.get("hierarchical_child_code", "0") == "0":
        context.patient_record = context.subscriber_record


@match("NM1", conditions={"entity_identifier_code": "IL"})
def set_subscriber_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the subscriber name loop 2010BA

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """

    if context.loop_name == TransactionLoops.SUBSCRIBER:

        if TransactionLoops.SUBSCRIBER_NAME not in context.subscriber_record:
            context.subscriber_record[TransactionLoops.SUBSCRIBER_NAME] = {
                "ref_segment": []
            }

        subscriber_name_loop = context.subscriber_record[
            TransactionLoops.SUBSCRIBER_NAME
        ]
        context.set_loop_context(TransactionLoops.SUBSCRIBER_NAME, subscriber_name_loop)


@match("NM1", conditions={"entity_identifier_code": "PR"})
def set_subscriber_payer_name_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the subscriber payer name loop 2010BB

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.SUBSCRIBER_NAME:

        if TransactionLoops.SUBSCRIBER_PAYER_NAME not in context.subscriber_record:
            context.subscriber_record[TransactionLoops.SUBSCRIBER_PAYER_NAME] = {
                "ref_segment": []
            }

        subscriber_payer_name_loop = context.subscriber_record[
            TransactionLoops.SUBSCRIBER_PAYER_NAME
        ]
        context.set_loop_context(
            TransactionLoops.SUBSCRIBER_NAME, subscriber_payer_name_loop
        )


@match("HL", conditions={"hierarchical_level_code": "23"})
def set_patient_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the patient loop 2000C

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if TransactionLoops.PATIENT not in context.subscriber_record:
        context.subscriber_record[TransactionLoops.PATIENT] = []

    context.subscriber_record[TransactionLoops.PATIENT].append({})
    context.patient_record = context.subscriber_record[TransactionLoops.PATIENT][-1]
    context.set_loop_context(TransactionLoops.PATIENT, context.patient_record)


@match("NM1", conditions={"entity_identifier_code": "QC"})
def set_patient_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the patient name loop 2010CA

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    context.patient_record[TransactionLoops.PATIENT_NAME] = {"ref_segment": []}
    patient_name_loop = context.patient_record[TransactionLoops.PATIENT_NAME]
    context.set_loop_context(TransactionLoops.PATIENT_NAME, patient_name_loop)


@match("CLM")
def set_claim_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the claim loop 2300

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if TransactionLoops.CLAIM_INFORMATION not in context.patient_record:
        context.patient_record[TransactionLoops.CLAIM_INFORMATION] = []

    context.patient_record[TransactionLoops.CLAIM_INFORMATION].append(
        {
            "dtp_segment": [],
            "pwk_segment": [],
            "ref_segment": [],
            "k3_segment": [],
            "hi_segment": [],
        }
    )

    claim_loop = context.patient_record[TransactionLoops.CLAIM_INFORMATION][-1]
    context.set_loop_context(TransactionLoops.CLAIM_INFORMATION, claim_loop)


@match(
    "NM1", conditions={"entity_identifier_code": ["71", "72", "ZZ", "82", "77", "DN"]}
)
def set_claim_entity_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the "entity" loops within the 2300 claims loop, including:
    * Attending Provider Name - Loop 2310A
    * Operating Physician Name - Loop 2310B
    * Other Operating Physician Name Name - Loop 2310C
    * Rendering Provider Name - Loop 2310D
    * Service Facility Location Name - Loop 2310E
    * Referring Provider Name - Loop 2310F

    """
    if "loop_2300" in context.loop_name or "loop_2310" in context.loop_name:
        identifier = segment_data.get("entity_identifier_code")
        loop_name = None

        if identifier == "71":
            loop_name = TransactionLoops.CLAIM_ATTENDING_PROVIDER_NAME
        elif identifier == "72":
            loop_name = TransactionLoops.CLAIM_OPERATING_PHYSICIAN_NAME
        elif identifier == "ZZ":
            loop_name = TransactionLoops.CLAIM_OTHER_OPERATING_PHYSICIAN_NAME
        elif identifier == "82":
            loop_name = TransactionLoops.CLAIM_RENDERING_PROVIDER_NAME
        elif identifier == "77":
            loop_name = TransactionLoops.CLAIM_SERVICE_FACILITY_LOCATION_NAME
        elif identifier == "DN":
            loop_name = TransactionLoops.CLAIM_REFERRING_PROVIDER_NAME

        if not loop_name:
            raise ValueError(
                f"unable to map identifier {identifier} to a claim entity loop"
            )

        claim = _get_claim(context)
        if loop_name not in claim:
            claim[loop_name] = {"ref_segment": []}

        loop_record = claim[loop_name]
        context.set_loop_context(loop_name, loop_record)


@match("SBR")
def set_other_subscriber_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the claim other subscriber loop 2320

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if "loop_23" in context.loop_name:
        claim = _get_claim(context)
        if TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION not in claim:
            claim[TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION] = []

        claim[TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION].append(
            {
                "cas_segment": [],
                "amt_segment": [],
            }
        )
        other_subscriber_loop = claim[
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION
        ][-1]
        context.set_loop_context(
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION, other_subscriber_loop
        )


@match("NM1", conditions={"entity_identifier_code": "IL"})
def set_other_subscriber_name_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the claim other subscriber name loop 2330A

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.CLAIM_OTHER_SUBSCRIBER_INFORMATION:
        other_subscriber = _get_other_subscriber(context)
        other_subscriber[TransactionLoops.CLAIM_OTHER_SUBSCRIBER_NAME] = {
            "ref_segment": []
        }

        other_subscriber_name = other_subscriber[
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_NAME
        ]
        context.set_loop_context(
            TransactionLoops.CLAIM_OTHER_SUBSCRIBER_NAME, other_subscriber_name
        )


@match(
    "NM1",
    conditions={
        "entity_identifier_code": ["IL", "PR", "71", "72", "ZZ", "77", "82", "DN", "85"]
    },
)
def set_other_subscriber_entities_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the other subscriber entities within the claim 2300 loop:
    * Other Subscriber Name - Loop 2330A
    * Other Payer Name - Loop 2330B
    * Other Payer Attending Provider - Loop 2330C
    * Other Payer Operating Physician - Loop 2330D
    * Other Payer Other Operating Physician - 2330E
    * Other Payer Service Facility Location - 2330F
    * Other Payer Rendering Provider Name - 2330G
    * Other Payer Referring Provider Name - 2330H
    * Other Payer Billing Provider Name - 2330I
    """
    if "loop_2320" in context.loop_name or "loop_2330" in context.loop_name:
        identifier = segment_data.get("entity_identifier_code")
        loop_name = None

        if identifier == "IL":
            loop_name = TransactionLoops.CLAIM_OTHER_SUBSCRIBER_NAME
        elif identifier == "PR":
            loop_name = TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_NAME
        elif identifier == "71":
            loop_name = (
                TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_ATTENDING_PROVIDER_NAME
            )
        elif identifier == "72":
            loop_name = (
                TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_OPERATING_PHYSICIAN_NAME
            )
        elif identifier == "ZZ":
            loop_name = (
                TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_OTHER_OPERATING_PHYSICIAN
            )
        elif identifier == "77":
            loop_name = (
                TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_SERVICE_FACILITY_LOCATION
            )
        elif identifier == "82":
            loop_name = (
                TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_RENDERING_PROVIDER_NAME
            )
        elif identifier == "DN":
            loop_name = (
                TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_REFERRING_PROVIDER_NAME
            )
        elif identifier == "85":
            loop_name = (
                TransactionLoops.CLAIM_OTHER_SUBSCRIBER_OTHER_PAYER_BILLING_PROVIDER_NAME
            )

        if not loop_name:
            raise ValueError(
                f"unable to map identifier {identifier} to a claim entity loop"
            )

        other_subscriber = _get_other_subscriber(context)

        if identifier == "IL":
            other_subscriber[loop_name] = {}
        else:
            other_subscriber[loop_name] = {"ref_segment": []}

        context.set_loop_context(loop_name, other_subscriber[loop_name])


@match("LX")
def set_service_line_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the claim service line loop 2400

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    claim = _get_claim(context)

    if TransactionLoops.CLAIM_SERVICE_LINE not in claim:
        claim[TransactionLoops.CLAIM_SERVICE_LINE] = []

    claim[TransactionLoops.CLAIM_SERVICE_LINE].append(
        {
            "pwk_segment": [],
            "ref_segment": [],
            "amt_segment": [],
        }
    )

    service_line_loop = claim[TransactionLoops.CLAIM_SERVICE_LINE][-1]
    context.set_loop_context(TransactionLoops.CLAIM_SERVICE_LINE, service_line_loop)


@match("LIN")
def set_drug_identification_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the service line drug identification loop 2410

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if "loop_24" in context.loop_name:
        service_line = _get_service_line(context)
        service_line[TransactionLoops.CLAIM_SERVICE_LINE_DRUG_IDENTIFICATION_NAME] = {}
        loop_record = service_line[
            TransactionLoops.CLAIM_SERVICE_LINE_DRUG_IDENTIFICATION_NAME
        ]
        context.set_loop_context(
            TransactionLoops.CLAIM_SERVICE_LINE_DRUG_IDENTIFICATION_NAME, loop_record
        )


@match(
    "NM1",
    conditions={"entity_identifier_code": ["72", "ZZ", "82", "DN"]},
)
def set_service_line_entities_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the entities within the service line loop 2400:
    *

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if "loop_24" in context.loop_name:
        identifier = segment_data.get("entity_identifier_code")
        loop_name = None

        if identifier == "72":
            loop_name = TransactionLoops.CLAIM_SERVICE_LINE_OPERATING_PHYSICIAN_NAME
        elif identifier == "ZZ":
            loop_name = (
                TransactionLoops.CLAIM_SERVICE_LINE_OTHER_OPERATING_PHYSICIAN_NAME
            )
        elif identifier == "82":
            loop_name = TransactionLoops.CLAIM_SERVICE_LINE_RENDERING_PROVIDER_NAME
        elif identifier == "DN":
            loop_name = TransactionLoops.CLAIM_SERVICE_LINE_REFERRING_PROVIDER_NAME

        if not loop_name:
            raise ValueError(
                f"unable to map identifier {identifier} to a claim entity loop"
            )
        service_line = _get_service_line(context)
        service_line[loop_name] = {"ref_segment": []}
        loop_record = service_line[loop_name]
        context.set_loop_context(loop_name, loop_record)


@match("SVD")
def set_service_line_adjudication_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets service line adjudication loop 2430

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    service_line = _get_service_line(context)

    if TransactionLoops.CLAIM_SERVICE_LINE_LINE_ADJUDICATION_INFORMATION:
        service_line[
            TransactionLoops.CLAIM_SERVICE_LINE_LINE_ADJUDICATION_INFORMATION
        ] = []

    service_line[
        TransactionLoops.CLAIM_SERVICE_LINE_LINE_ADJUDICATION_INFORMATION
    ].append(
        {
            "cas_segment": [],
        }
    )

    adjudication_loop = service_line[
        TransactionLoops.CLAIM_SERVICE_LINE_LINE_ADJUDICATION_INFORMATION
    ][-1]
    context.set_loop_context(
        TransactionLoops.CLAIM_SERVICE_LINE_LINE_ADJUDICATION_INFORMATION,
        adjudication_loop,
    )


@match("SE")
def set_se_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set footer loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """

    context.set_loop_context(
        TransactionLoops.FOOTER, context.transaction_data["footer"]
    )



"""
loops.py

Models the loops, or logical segment groupings, for the Eligibility 270 005010X279A1 transaction set.
The Eligibility Transaction set organizes loops into a hierarchical and nested model.

- Header
- Loop 2000A (Information Source)
    -- Loop 2100A (Information Source Name)
    -- Loop 2000B (Information Receiver)
        --- Loop 2100B (Information Receiver Name)
        --- Loop 2000C (Subscriber)
            --- Loop 2100C (Subscriber Name)
                --- Loop 2110C (Subscriber Eligibility)
            --- Loop 2000D (Dependent)
                --- Loop 2100D (Dependent Name)
                    --- Loop 2110D (Dependent Eligibility)
-- Footer

The Header and Footer components are not "loops" per the specification, but are included to standardize and simplify
transactional modeling and processing.
"""

from x12.models import X12SegmentGroup
from x12.segments import (
    SeSegment,
    N3Segment,
    N4Segment,
    TrnSegment,
    DmgSegment,
    HiSegment,
)
from .segments import (
    HeaderStSegment,
    HeaderBhtSegment,
    Loop2000DHlSegment,
    Loop2000CHlSegment,
    Loop2100BNm1Segment,
    Loop2000BHlSegment,
    Loop2100BRefSegment,
    Loop2100ANm1Segment,
    Loop2000AHlSegment,
    Loop2100CNm1Segment,
    Loop2100RefSegment,
    Loop2100CInsSegment,
    Loop2100DtpSegment,
    Loop2110EqSegment,
    Loop2110AmtSegment,
    Loop2110IiiSegment,
    Loop2110RefSegment,
    Loop2110DtpSegment,
    Loop2100DNm1Segment,
    Loop2100DInsSegment,
    Loop2100CPrvSegment,
    Loop2100BPrvSegment,
)
from typing import List, Optional
from pydantic import Field, root_validator
from x12.validators import validate_duplicate_ref_codes


class Header(X12SegmentGroup):
    """
    Transaction Header Information
    """

    st_segment: HeaderStSegment
    bht_segment: HeaderBhtSegment


class Loop2110D(X12SegmentGroup):
    """
    Loop 2110D Dependent Eligibility
    """

    eq_segment: Optional[Loop2110EqSegment]
    amt_segment: Optional[List[Loop2110AmtSegment]] = Field(min_items=0, max_items=2)
    iii_segment: Optional[Loop2110IiiSegment]
    ref_segment: Optional[Loop2110RefSegment]
    dtp_segment: Optional[Loop2110DtpSegment]


class Loop2100D(X12SegmentGroup):
    """
    Loop 2100D - Dependent Name
    """

    nm1_segment: Loop2100DNm1Segment
    ref_segment: Optional[List[Loop2100RefSegment]] = Field(min_items=0, max_items=9)
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    prv_segment: Optional[Loop2100CPrvSegment]
    dmg_segment: Optional[DmgSegment]
    ins_segment: Optional[Loop2100DInsSegment]
    hi_segment: Optional[HiSegment]
    dtp_segment: Optional[Loop2100DtpSegment]
    loop_2110d: Loop2110D

    _validate_ref_segments = root_validator(allow_reuse=True)(
        validate_duplicate_ref_codes
    )


class Loop2000D(X12SegmentGroup):
    """
    Loop 2000D - Dependent
    """

    hl_segment: Loop2000DHlSegment
    trn_segment: Optional[List[TrnSegment]] = Field(min_items=0, max_items=2)
    loop_2100d: Loop2100D


class Loop2110C(X12SegmentGroup):
    """
    Loop2110C - Subscriber Eligibility
    """

    eq_segment: Optional[Loop2110EqSegment]
    amt_segment: Optional[List[Loop2110AmtSegment]] = Field(min_items=0, max_items=2)
    iii_segment: Optional[Loop2110IiiSegment]
    ref_segment: Optional[Loop2110RefSegment]
    dtp_segment: Optional[Loop2110DtpSegment]


class Loop2100C(X12SegmentGroup):
    """
    Loop 2100C - Subscriber Name
    """

    nm1_segment: Loop2100CNm1Segment
    ref_segment: Optional[List[Loop2100RefSegment]] = Field(min_items=0, max_items=9)
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    prv_segment: Optional[Loop2100CPrvSegment]
    dmg_segment: Optional[DmgSegment]
    ins_segment: Optional[Loop2100CInsSegment]
    hi_segment: Optional[HiSegment]
    dtp_segment: Optional[Loop2100DtpSegment]
    loop_2110c: Optional[Loop2110C]

    _validate_ref_segments = root_validator(allow_reuse=True)(
        validate_duplicate_ref_codes
    )


class Loop2000C(X12SegmentGroup):
    """
    Loop 2000C - Subscriber
    """

    hl_segment: Loop2000CHlSegment
    trn_segment: Optional[List[TrnSegment]] = Field(min_items=0, max_items=2)
    loop_2100c: Loop2100C
    loop_2000d: Optional[List[Loop2000D]] = Field(min_items=0)


class Loop2100B(X12SegmentGroup):
    """
    Loop 2100B - Information Receiver Name
    """

    nm1_segment: Loop2100BNm1Segment
    ref_segment: Optional[List[Loop2100BRefSegment]]
    n3_segment: Optional[N3Segment]
    n4_segment: Optional[N4Segment]
    prv_segment: Optional[Loop2100BPrvSegment]

    _validate_ref_segments = root_validator(allow_reuse=True)(
        validate_duplicate_ref_codes
    )


class Loop2000B(X12SegmentGroup):
    """
    Loop 2000B - Information Receiver
    """

    hl_segment: Loop2000BHlSegment
    loop_2100b: Loop2100B
    loop_2000c: List[Loop2000C] = Field(min_items=1)


class Loop2100A(X12SegmentGroup):
    """
    Loop 2100A - Information Source Name
    """

    nm1_segment: Loop2100ANm1Segment


class Loop2000A(X12SegmentGroup):
    """
    Loop 2000A - Information Source
    The root node/loop for the 270 transaction

    """

    hl_segment: Loop2000AHlSegment
    loop_2100a: Loop2100A
    loop_2000b: List[Loop2000B] = Field(min_items=1)


class Footer(X12SegmentGroup):
    """
    Transaction Footer Information
    """

    se_segment: SeSegment



"""
Tests use-cases for the 270 005010X279A1 (Eligibility Inquiry) transaction
"""

import pytest
from tests.support import assert_eq_model, resources_directory
import os


@pytest.fixture
def resource_path() -> str:
    return os.path.join(resources_directory, "270_005010X279A1")


@pytest.mark.parametrize(
    "file_name",
    ["dependent-health-benefit-check.270", "subscriber-health-benefit-check.270"],
)
def test_270_model(resource_path, file_name: str):

    x12_file_path = os.path.join(resource_path, file_name)
    assert os.path.exists(x12_file_path)
    assert_eq_model(x12_file_path)



"""
test_837_005010XX222A2.py
"""
import pytest
from tests.support import assert_eq_model, resources_directory
import os


@pytest.fixture
def resource_path() -> str:
    return os.path.join(resources_directory, "837_005010X222A2")


@pytest.mark.parametrize(
    "file_name",
    [
        "demo.ambulance.example5.837",
        "demo.autoaccident.837",
        "demo.cob.2ndary.example4.837",
        "demo.cob.example3.A.837",
        "demo.cob.example3.B.837",
        "demo.cob.example3.C.837",
        "demo.cob.example4.837",
        "demo.drug.example10.1.837",
        "demo.drug.example10.2.837",
        "demo.drug.example10.3.837",
        "demo.example6.837",
        "demo.example7.837",
        "demo.example8.837",
        "demo.example9.837",
        "demo.example11.837",
        "demo.example12.837",
        "demo.example1.837",
        "demo.example2.837",
    ],
)
def test_837_model(resource_path, file_name: str):
    x12_file_path = os.path.join(resource_path, file_name)
    assert os.path.exists(x12_file_path)
    assert_eq_model(x12_file_path)



"""
parsing.py

Parses X12 276 005010X212 segments into a transactional domain model.

The parsing module includes a specific parser for the 276 transaction and loop parsing functions to create new loops
as segments are streamed to the transactional data model.

Loop parsing functions are implemented as set_[description]_loop(context: X12ParserContext, segment_data: Dict).
"""
from enum import Enum
from x12.parsing import X12ParserContext, match
from typing import Dict, Optional


class TransactionLoops(str, Enum):
    """
    The loops used to support the 276 005010X212 format.
    """

    HEADER = "header"
    INFORMATION_SOURCE_LEVEL = "loop_2000a"
    PAYER_NAME = "loop_2100a"
    INFORMATION_RECEIVER_LEVEL = "loop_2000b"
    INFORMATION_RECEIVER_NAME = "loop_2100b"
    SERVICE_PROVIDER_LEVEL = "loop_2000c"
    SERVICE_PROVIDER_NAME = "loop_2100c"
    SUBSCRIBER_LEVEL = "loop_2000d"
    SUBSCRIBER_NAME = "loop_2100d"
    SUBSCRIBER_CLAIM_STATUS_TRACKING_NUMBER = "loop_2200d"
    SUBSCRIBER_SERVICE_LINE_INFORMATION = "loop_2210d"
    DEPENDENT_LEVEL = "loop_2000e"
    DEPENDENT_NAME = "loop_2100e"
    DEPENDENT_CLAIM_STATUS_TRACKING_NUMBER = "loop_2200e"
    DEPENDENT_SERVICE_LINE_INFORMATION = "loop_2210e"
    FOOTER = "footer"


def _get_information_source(context: X12ParserContext) -> Optional[Dict]:
    return context.transaction_data[TransactionLoops.INFORMATION_SOURCE_LEVEL][-1]


def _get_information_receiver(context: X12ParserContext) -> Dict:
    """Returns the current information receiver record"""
    information_source = _get_information_source(context)
    return information_source[TransactionLoops.INFORMATION_RECEIVER_LEVEL][-1]


def _get_service_provider(
    context: X12ParserContext, hierarchical_id: str
) -> Optional[Dict]:
    """
    Returns the service provider record by id.

    :param hierarchical_id: The service provider hierarchical level id
    """

    information_receiver = _get_information_receiver(context)

    for sp in information_receiver.get(TransactionLoops.SERVICE_PROVIDER_LEVEL, []):
        sp_hierarchical_id = sp.get("hl_segment", {}).get("hierarchical_id_number")
        if sp_hierarchical_id == hierarchical_id:
            return sp

    return None


def _is_subscriber(patient_record: Dict):
    """Returns true if the patient is a subscriber record"""
    return patient_record.get("hl_segment", {}).get("hierarchical_level_code") == "22"


@match("ST")
def set_header_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set header loop for the 276 transaction set.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """

    context.set_loop_context(
        TransactionLoops.HEADER, context.transaction_data[TransactionLoops.HEADER]
    )


@match("HL", {"hierarchical_level_code": "20"})
def set_information_source_loop(context: X12ParserContext, segment_data: Dict):
    """
    Sets the information source loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if TransactionLoops.INFORMATION_SOURCE_LEVEL not in context.transaction_data:
        context.transaction_data[TransactionLoops.INFORMATION_SOURCE_LEVEL] = []

    context.transaction_data[TransactionLoops.INFORMATION_SOURCE_LEVEL].append({})
    loop_record = context.transaction_data[TransactionLoops.INFORMATION_SOURCE_LEVEL][
        -1
    ]
    context.set_loop_context(TransactionLoops.INFORMATION_SOURCE_LEVEL, loop_record)


@match("NM1", {"entity_identifier_code": "PR"})
def set_payer_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the Information Source/Payer Name Loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.INFORMATION_SOURCE_LEVEL:
        information_source = _get_information_source(context)
        information_source[TransactionLoops.PAYER_NAME] = {}
        loop_record = information_source[TransactionLoops.PAYER_NAME]
        context.set_loop_context(TransactionLoops.PAYER_NAME, loop_record)


@match("HL", {"hierarchical_level_code": "21"})
def set_information_receiver_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the information receiver loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    information_source = _get_information_source(context)
    if TransactionLoops.INFORMATION_RECEIVER_LEVEL not in information_source:
        information_source[TransactionLoops.INFORMATION_RECEIVER_LEVEL] = []

    information_source[TransactionLoops.INFORMATION_RECEIVER_LEVEL].append({})
    loop_record = information_source[TransactionLoops.INFORMATION_RECEIVER_LEVEL][-1]
    context.set_loop_context(TransactionLoops.INFORMATION_RECEIVER_LEVEL, loop_record)


@match("NM1", {"entity_identifier_code": "41"})
def set_information_receiver_name_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the Information Receiver Name Loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.INFORMATION_RECEIVER_LEVEL:
        information_receiver = _get_information_receiver(context)
        information_receiver[TransactionLoops.INFORMATION_RECEIVER_NAME] = {}
        loop_record = information_receiver[TransactionLoops.INFORMATION_RECEIVER_NAME]
        context.set_loop_context(
            TransactionLoops.INFORMATION_RECEIVER_NAME, loop_record
        )


@match("HL", {"hierarchical_level_code": "19"})
def set_service_provider_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the service provider loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    information_receiver = _get_information_receiver(context)
    if TransactionLoops.SERVICE_PROVIDER_LEVEL not in information_receiver:
        information_receiver[TransactionLoops.SERVICE_PROVIDER_LEVEL] = []

    information_receiver[TransactionLoops.SERVICE_PROVIDER_LEVEL].append({})
    loop_record = information_receiver[TransactionLoops.SERVICE_PROVIDER_LEVEL][-1]
    context.set_loop_context(TransactionLoops.SERVICE_PROVIDER_LEVEL, loop_record)


@match("NM1", {"entity_identifier_code": "1P"})
def set_service_provider_name_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the Service Provider Name Loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.SERVICE_PROVIDER_LEVEL:
        context.loop_container[TransactionLoops.SERVICE_PROVIDER_NAME] = {}
        loop_record = context.loop_container[TransactionLoops.SERVICE_PROVIDER_NAME]
        context.set_loop_context(TransactionLoops.SERVICE_PROVIDER_NAME, loop_record)


@match("HL", {"hierarchical_level_code": "22"})
def set_subscriber_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the subscriber loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    service_provider_id = segment_data.get("hierarchical_parent_id_number")
    service_provider = _get_service_provider(context, service_provider_id)

    if service_provider is None:
        raise LookupError(f"Service Provider ID {service_provider_id} not found")

    if TransactionLoops.SUBSCRIBER_LEVEL not in service_provider:
        service_provider[TransactionLoops.SUBSCRIBER_LEVEL] = []

    service_provider[TransactionLoops.SUBSCRIBER_LEVEL].append({})
    context.subscriber_record = service_provider[TransactionLoops.SUBSCRIBER_LEVEL][-1]
    context.set_loop_context(
        TransactionLoops.SUBSCRIBER_LEVEL, context.subscriber_record
    )
    # subscriber is patient
    if segment_data.get("hierarchical_child_code", "0") == "0":
        context.patient_record = context.subscriber_record


@match("NM1", {"entity_identifier_code": "IL"})
def set_subscriber_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the Subscriber Name Loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.SUBSCRIBER_LEVEL:
        context.subscriber_record[TransactionLoops.SUBSCRIBER_NAME] = {}
        loop_record = context.subscriber_record[TransactionLoops.SUBSCRIBER_NAME]
        context.set_loop_context(TransactionLoops.SUBSCRIBER_NAME, loop_record)


@match("TRN")
def set_claim_status_tracking_loop(
    context: X12ParserContext, segment_data: Dict
) -> None:
    """
    Sets the claim status tracking loop for the patient (subscriber or dependent)

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if _is_subscriber(context.patient_record):
        loop_name = TransactionLoops.SUBSCRIBER_CLAIM_STATUS_TRACKING_NUMBER
    else:
        loop_name = TransactionLoops.DEPENDENT_CLAIM_STATUS_TRACKING_NUMBER

    if loop_name not in context.patient_record:
        context.patient_record[loop_name] = []

    context.patient_record[loop_name].append({"ref_segment": []})
    loop_record = context.patient_record[loop_name][-1]
    context.set_loop_context(loop_name, loop_record)


@match("SVC")
def set_service_line_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the service line loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    claim_status = context.loop_container

    if _is_subscriber(context.patient_record):
        loop_name = TransactionLoops.SUBSCRIBER_SERVICE_LINE_INFORMATION
    else:
        loop_name = TransactionLoops.DEPENDENT_SERVICE_LINE_INFORMATION

    if loop_name not in claim_status:
        claim_status[loop_name] = []

    claim_status[loop_name].append({})
    loop_record = claim_status[loop_name][-1]
    context.set_loop_context(loop_name, loop_record)


@match("HL", {"hierarchical_level_code": "23"})
def set_dependent_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the dependent loop
    """
    if TransactionLoops.DEPENDENT_LEVEL not in context.subscriber_record:
        context.subscriber_record[TransactionLoops.DEPENDENT_LEVEL] = []

    context.subscriber_record[TransactionLoops.DEPENDENT_LEVEL].append({})
    context.patient_record = context.subscriber_record[
        TransactionLoops.DEPENDENT_LEVEL
    ][-1]
    context.set_loop_context(TransactionLoops.DEPENDENT_LEVEL, context.patient_record)


@match("NM1", {"entity_identifier_code": "QC"})
def set_dependent_name_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the Dependent Name Loop

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """
    if context.loop_name == TransactionLoops.DEPENDENT_LEVEL:
        context.patient_record[TransactionLoops.DEPENDENT_NAME] = {}
        loop_record = context.patient_record[TransactionLoops.DEPENDENT_NAME]
        context.set_loop_context(TransactionLoops.DEPENDENT_NAME, loop_record)


@match("SE")
def set_footer_loop(context: X12ParserContext, segment_data: Dict) -> None:
    """
    Sets the transaction set footer loop.

    :param context: The X12Parsing context which contains the current loop and transaction record.
    :param segment_data: The current segment data
    """

    context.set_loop_context(
        TransactionLoops.FOOTER, context.transaction_data["footer"]
    )



"""
segments.py

Specialized segment models for the Health Care Claim Status Request 276 005010X212 transaction.
"""
from x12.segments import (
    StSegment,
    BhtSegment,
    HlSegment,
    Nm1Segment,
    DmgSegment,
    TrnSegment,
    RefSegment,
    AmtSegment,
    DtpSegment,
    SvcSegment,
)
from typing import Literal, Optional, Union
from enum import Enum
import datetime
from decimal import Decimal


class HeaderStSegment(StSegment):
    """
    Customized ST segment for 837 transaction header
    """

    transaction_set_identifier_code: Literal["276"]
    implementation_convention_reference: Literal["005010X212"]


class HeaderBhtSegment(BhtSegment):
    """
    Header BHT Segment
    """

    hierarchical_structure_code: Literal["0010"]
    transaction_set_purpose_code: Literal["13"]
    transaction_type_code: Optional[str]


class Loop2000AHlSegment(HlSegment):
    """
    Information Source HL Segment
    """

    hierarchical_level_code: Literal["20"]
    hierarchical_parent_id_number: Optional[str]
    hierarchical_child_code: Literal["1"]


class Loop2100ANm1Segment(Nm1Segment):
    """
    Payer Name
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        PAYOR_IDENTIFICATION = "PI"
        CMS_PLAN_ID = "XV"

    entity_identifier_code: Literal["PR"]
    entity_type_qualifier: Literal["2"]
    name_first: Optional[str]
    identification_code_qualifier: IdentificationCodeQualifier


class Loop2000BHlSegment(HlSegment):
    """
    Information Receiver HL Segment
    """

    hierarchical_level_code: Literal["21"]
    hierarchical_child_code: Literal["1"]


class Loop2100BNm1Segment(Nm1Segment):
    """
    Information Receiver Name Segment
    """

    entity_identifier_code: Literal["41"]
    identification_code_qualifier: Literal["46"]


class Loop2000CHlSegment(HlSegment):
    """
    Service Provider HL Segment
    """

    hierarchical_level_code: Literal["19"]
    hierarchical_child_code: Literal["1"]


class Loop2100CNm1Segment(Nm1Segment):
    """
    Service Provider NM1 Segment
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        FEDERAL_TAXPAYER_IDENTIFIER_NUMBER = "FI"
        SERVICE_PROVIDER_NUMBER = "SV"
        CMS_NATIONAL_PROVIDER_IDENTIFIER = "XX"

    entity_identifier_code: Literal["1P"]
    name_last_or_organization_name: Optional[str]
    identification_code_qualifier: IdentificationCodeQualifier


class Loop2000DHlSegment(HlSegment):
    """
    Subscriber HL Segment
    """

    hierarchical_level_code: Literal["22"]


class Loop2000DDmgSegment(DmgSegment):
    """
    Subscriber Demographics
    """

    date_time_period_format_qualifier: Literal["D8"]
    date_time_period: Optional[Union[str, datetime.date]]


class Loop2100DNm1Segment(Nm1Segment):
    """
    Subscriber Name
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for NM108
        """

        EMPLOYER_IDENTIFICATION_NUMBER = "24"
        STANDARD_US_HEALTH_IDENTIFIER = "II"
        MEMBER_IDENTIFICATION_NUMBER = "MI"

    entity_identifier_code: Literal["IL"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: IdentificationCodeQualifier


class Loop2200DTrnSegment(TrnSegment):
    """
    Subscriber Claim Status Tracking Trace Segment
    """

    trace_type_code: Literal["1"]
    originating_company_identifier: Optional[str]
    reference_identification_2: Optional[str]


class Loop2200DRefSegment(RefSegment):
    """
    Subscriber Claim Status Tracking Trace Segment Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PAYOR_CLAIM_NUMBER = "1K"
        BILLING_TYPE = "BLT"
        LOCATION_NUMBER = "LU"
        GROUP_NUMBER = "6P"
        PATIENT_ACCOUNT_NUMBER = "EJ"
        PHARMACY_PRESCRIPTION_NUMBER = "XZ"
        CLAIM_NUMBER = "D9"


class Loop2200DAmtSegment(AmtSegment):
    """
    Subscriber Claim Status Tracking Amount Segment
    """

    amount_qualifier_code: Literal["T3"]


class Loop2200DDtpSegment(DtpSegment):
    """
    Subscriber Claim Status Tracking Amount Segment Dates
    """

    date_time_qualifier: Literal["472"]


class Loop2210DSvcSegment(SvcSegment):
    """
    Subscriber Claim Status Service Line
    """

    line_item_provider_payment_amount: Optional[Decimal]
    original_units_of_service_count: Decimal


class Loop2210DRefSegment(RefSegment):
    """
    Subscriber Claim Status Service Line Reference Identification
    """

    reference_identification_qualifier: Literal["FJ"]


class Loop2210DDtpSegment(DtpSegment):
    """
    Subscriber Claim Status Service Line Dates
    """

    date_time_qualifier: Literal["472"]


class Loop2000EHlSegment(HlSegment):
    """
    Dependent HL Segment
    """

    hierarchical_level_code: Literal["23"]
    hierarchical_child_code: Optional[str]


class Loop2100ENm1Segment(Nm1Segment):
    """
    Dependent Name Segment
    """

    entity_identifier_code: Literal["QC"]
    entity_type_qualifier: Literal["1"]
    identification_code_qualifier: Optional[str]
    identification_code: Optional[str]



"""
test_x12_276_005010X212.py

Tests the Health Care Claim Status Request Transaction
"""
import pytest
from tests.support import assert_eq_model, resources_directory
import os


@pytest.fixture
def resource_path() -> str:
    return os.path.join(resources_directory, "276_005010X212")


@pytest.mark.parametrize(
    "file_name",
    [
        "claim-level-status.276",
        "claim-level-status-ncpdp.276",
        "information-receiver-level-status.276",
        "provider-level-status.276",
    ],
)
def test_276_model(resource_path, file_name: str):
    x12_file_path = os.path.join(resource_path, file_name)
    assert os.path.exists(x12_file_path)
    assert_eq_model(x12_file_path)



import pytest
from typing import Callable, Dict
from x12.parsing import X12ParserContext, match


@pytest.fixture
def segment_data(x12_delimiters) -> Dict:
    return {
        "delimiters": x12_delimiters.dict(),
        "segment_name": "ST",
        "transaction_set_identifier_code": "270",
        "transaction_set_control_number": "0001",
        "implementation_convention_reference": "005010X279A1",
    }


@pytest.fixture
def parser_function() -> Callable:
    def func(x12_parser_context: X12ParserContext, segment_data: Dict):
        x12_parser_context.parsed_loops.append("test_loop")

    return func


def test_match_no_conditions(parser_function, segment_data, x12_parser_context):
    """
    Tests the match decorator when no conditions are specified.
    """
    wrapped_func = match("ST")(parser_function)
    wrapped_func(segment_data, x12_parser_context)
    assert x12_parser_context.loop_name == "test_loop"


def test_single_match_condition(parser_function, segment_data, x12_parser_context):
    """
    Tests the match decorator when a single condition is specified.
    """
    wrapped_func = match("ST", conditions={"transaction_set_identifier_code": "270"})(
        parser_function
    )
    wrapped_func(segment_data, x12_parser_context)
    assert x12_parser_context.loop_name == "test_loop"


def test_multiple_match_condition(parser_function, segment_data, x12_parser_context):
    """
    Tests the match decorator when multiple conditions are specified.
    """
    wrapped_func = match(
        "ST",
        conditions={
            "transaction_set_identifier_code": "270",
            "transaction_set_control_number": "0001",
        },
    )(parser_function)
    wrapped_func(segment_data, x12_parser_context)
    assert x12_parser_context.loop_name == "test_loop"


def test_list_match_condition(parser_function, segment_data, x12_parser_context):
    """
    Tests the match decorator when a list/"one of" condition is specified.
    """
    wrapped_func = match(
        "ST", conditions={"transaction_set_identifier_code": ["270", "276"]}
    )(parser_function)
    wrapped_func(segment_data, x12_parser_context)
    assert x12_parser_context.loop_name == "test_loop"


def test_match_conditions_unmatched(parser_function, segment_data, x12_parser_context):
    """
    Tests the match decorator when conditions are specified but they do not match the data_segment.
    """
    wrapped_func = match("ST", conditions={"transaction_set_identifier_code": "911"})(
        parser_function
    )
    wrapped_func(segment_data, x12_parser_context)
    assert x12_parser_context.loop_name is None



"""
transaction_set.py

Defines the Institutional Claims 837 005010X223A3 transaction set model.
"""

from x12.models import X12SegmentGroup
from .loops import Header, Footer, Loop1000A, Loop1000B, Loop2000A, Loop2300
from pydantic import Field, root_validator
from typing import List, Dict
from x12.validators import validate_segment_count


class HealthCareClaimInstitutional(X12SegmentGroup):
    """
    The HealthCare Claim - Institutional transaction model (837)
    """

    header: Header
    loop_1000a: Loop1000A
    loop_1000b: Loop1000B
    loop_2000a: List[Loop2000A] = Field(min_items=1)
    footer: Footer

    _validate_segment_count = root_validator(allow_reuse=True)(validate_segment_count)



"""
segments.py

Specialized segment models for the Health Care Claim Payment 835 005010X221A1 transaction set model.
"""
from x12.segments import (
    StSegment,
    TrnSegment,
    CurSegment,
    RefSegment,
    DtmSegment,
    PerSegment,
    N1Segment,
    CasSegment,
    AmtSegment,
    QtySegment,
    LqSegment,
)
from typing import Literal, Optional
from enum import Enum


class HeaderStSegment(StSegment):
    transaction_set_identifier_code: Literal["835"]
    implementation_convention_reference: Optional[str]


class HeaderTrnSegment(TrnSegment):
    """
    Header Trace Segment
    """

    trace_type_code: Literal["1"]


class HeaderCurSegment(CurSegment):
    """
    Header Currency Segment
    """

    entity_identifier_code: Literal["PR"]


class HeaderRefSegment(RefSegment):
    """
    Header Ref Segments
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        RECEIVER_IDENTIFICATION_NUMBER = "EV"
        VERSION_CODE_LOCAL = "F2"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class HeaderDtmSegment(DtmSegment):
    """
    Header Dtm Segment
    """

    date_time_qualifier: Literal["405"]


class Loop1000AN1Segment(N1Segment):
    """
    Payer Identification
    """

    entity_identifier_code: Literal["PR"]
    identification_code_qualifier: Optional[Literal["XV"]]


class Loop1000ARefSegment(RefSegment):
    """
    Loop 1000A Ref Segment
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        PAYER_IDENTIFICATION_NUMBER = "2U"
        SUBMITTER_IDENTIFICATION_NUMBER = "EO"
        HEALTH_INDUSTRY_NUMBER = "HI"
        NAIC_CODE = "NF"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop1000APerSegment(PerSegment):
    """
    Loop 1000A Per Segment
    """

    class ContactFunctionCode(str, Enum):
        """
        Code values for PER01
        """

        PAYERS_CLAIM_OFFICE = "CX"
        TECHNICAL_DEPARTMENT = "BL"
        INFORMATION_CONTACT = "IC"

    contact_function_code: ContactFunctionCode


class Loop1000BN1Segment(N1Segment):
    """
    Payee Identification
    """

    class IdentificationCodeQualifier(str, Enum):
        """
        Code values for N1003
        """

        FEDERAL_TAXPAYER_IDENTIFICATION_NUMBER = "FI"
        CMS_PLAN_ID = "XV"
        CMS_NPI = "XX"

    entity_identifier_code: Literal["PE"]
    identification_code_qualifier: IdentificationCodeQualifier


class Loop1000BRefSegment(RefSegment):
    """
    Payee Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        STATE_LICENSE_NUMBER = "0B"
        NATIONAL_COUNCIL_PHARMACY_NUMBER = "D3"
        PAYEE_IDENTIFICATION = "PQ"
        FEDERAL_TAXPAYERS_IDENTIFICATION_NUMBER = "TJ"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2100CasSegment(CasSegment):
    """
    Claim Payment Information CAS segment
    """

    class ClaimAdjustmentGroupCode(str, Enum):
        """
        Code values for CAS01
        """

        CONTRACTUAL_OBLIGATIONS = "CO"
        OTHER_ADJUSTMENTS = "OA"
        PAYER_INITIATED_REDUCTIONS = "PI"
        PATIENT_RESPONSIBILITY = "PR"

    adjustment_group_code: ClaimAdjustmentGroupCode


class Loop2100RefSegment(RefSegment):
    """
    Claim Payment Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        GROUP_POLICY_NUMBER = "1L"
        MEMBER_IDENTIFICATION_NUMBER = "1W"
        EMPLOYEE_IDENTIFICATION_NUMBER = "28"
        GROUP_NUMBER = "6P"
        REPRICED_CLAIM_REFERENCE_NUMBER = "9A"
        ADJUSTED_REPRICED_CLAIM_REFERENCE_NUMBER = "9C"
        AUTHORIZATION_NUMBER = "BB"
        CLASS_OF_CONTRACT_CODE = "CE"
        MEDICAL_RECORD_IDENTIFICATION_NUMBER = "EA"
        ORIGINAL_REFERENCE_NUMBER = "F8"
        PRIOR_AUTHORIZATION_NUMBER = "G1"
        PREDETERMINATION_BENEFITS_IDENTIFICATION_NUMBER = "G3"
        INSURANCE_POLICY_NUMBER = "IG"
        SOCIAL_SECURITY_NUMBER = "SY"
        STATE_LICENSE_NUMBER = "0B"
        BLUE_CROSS_PROVIDER_NUMBER = "1A"
        BLUE_SHIELD_PROVIDER_NUMBER = "1B"
        MEDICARE_PROVIDER_NUMBER = "1C"
        MEDICAID_PROVIDER_NUMBER = "1D"
        PROVIDER_UPIN_NUMBER = "1G"
        CHAMPUS_IDENTIFICATION_NUMBER = "1H"
        FACILITY_ID_NUMBER = "1J"
        NATIONAL_COUNCIL_PRESCRIPTION_PHARMACY_NUMBER = "D3"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        LOCATION_NUMBER = "LU"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2100DtmSegment(DtmSegment):
    """
    Claim Payment Information Date Time Reference
    """

    class DateTimeQualifier(str, Enum):
        """
        DTM01 code values
        """

        CLAIM_STATEMENT_PERIOD_START = "232"
        CLAIM_STATEMENT_PERIOD_STOP = "233"
        EXPIRATION = "036"
        RECEIVED = "050"

    date_time_qualifier: DateTimeQualifier


class Loop2100PerSegment(PerSegment):
    """
    Claim Payment Contact Information
    """

    contact_function_code: Literal["CX"]


class Loop2100AmtSegment(AmtSegment):
    """
    Claim Payment Amounts
    """

    class AmountQualifierCode(str, Enum):
        """
        Code values for AMT01
        """

        COVERAGE_AMOUNT = "AU"
        DISCOUNT_AMOUNT = "D8"
        PER_DAY_LIMIT = "DY"
        PATIENT_AMOUNT_PAID = "F5"
        INTEREST = "I"
        NEGATIVE_LEDGER_BALANCE = "NL"
        TAX = "T"
        TOTAL_CLAIM_BEFORE_TAXES = "T2"
        FEDERAL_MANDATE_CATEGORY_1 = "ZK"
        FEDERAL_MANDATE_CATEGORY_2 = "ZL"
        FEDERAL_MANDATE_CATEGORY_3 = "ZM"
        FEDERAL_MANDATE_CATEGORY_4 = "ZN"
        FEDERAL_MANDATE_CATEGORY_5 = "ZO"

    amount_qualifier_code: AmountQualifierCode


class Loop2100QtySegment(QtySegment):
    """
    Payment Information Quantity
    """

    class QuantityQualifier(str, Enum):
        """
        Code values for QTY01
        """

        COVERED_ACTUAL = "CA"
        COINSURED_ACTUAL = "CD"
        LIFE_TERM_RESERVE_ACTUAL = "LA"
        LIFE_TERM_RESERVE_ESTIMATED = "LE"
        NONCOVERED_ESTIMATED = "NE"
        NOT_REPLACED_BLOOD_UNITS = "NR"
        OUTLIER_DAYS = "OU"
        PRESCRIPTION = "PS"
        VISITS = "VS"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_1 = "ZK"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_2 = "ZL"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_3 = "ZM"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_4 = "ZN"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_5 = "ZO"

    quantity_qualifier: QuantityQualifier


class Loop2110DtmSegment(DtmSegment):
    """
    Service Line Payment Information Date
    """

    class DateTimeQualifier(str, Enum):
        """
        Code values for DTM01
        """

        SERVICE_PERIOD_START = "150"
        SERVICE_PERIOD_END = "151"
        SERVICE = "472"

    date_time_qualifier: DateTimeQualifier


class Loop2110CasSegment(CasSegment):
    """
    Service Line Payment Information Adjustments
    """

    class ClaimAdjustmentGroupCode(str, Enum):
        """
        Code values for CAS01
        """

        CONTRACTURAL_OBLIGATIONS = "CO"
        OTHER_ADJUSTMENTS = "OA"
        PAYOR_INITIATED_REDUCTIONS = "PI"
        PATIENT_RESPONSIBILITY = "PR"

    adjustment_group_code: ClaimAdjustmentGroupCode


class Loop2110RefSegment(RefSegment):
    """
    Service Line Payment Information Reference Identification
    """

    class ReferenceIdentificationQualifier(str, Enum):
        """
        Code values for REF01
        """

        AMBULATORY_PATIENT_GROUP_NUMBER = "1S"
        AMBULATORY_PAYMENT_CLASSIFICATION = "APC"
        AUTHORIZATION_NUMBER = "BB"
        ATTACHMENT_CODE = "E9"
        PRIOR_AUTHORIZATION_NUMBER = "G1"
        PREDETERMINATION_OF_BENEFITS_NUMBER = "G3"
        LOCATION_NUMBER = "LU"
        RATE_CODE_NUMBER = "RB"
        PROVIDER_CONTROL_NUMBER = "6R"
        STATE_LICENSE_NUMBER = "0B"
        BLUE_CROSS_PROVIDER_NUMBER = "1A"
        BLUE_SHIELD_PROVIDER_NUMBER = "1B"
        MEDICARE_PROVIDER_NUMBER = "1C"
        MEDICAID_PROVIDER_NUMBER = "1D"
        PROVIDER_UPIN_NUMBER = "1G"
        CHAMPUS_IDENTIFICATION_NUMBER = "1H"
        FACILITY_ID_NUMBER = "1J"
        NATIONAL_COUNCIL_PRESCRIPTION_PHARMACY_NUMBER = "D3"
        PROVIDER_COMMERCIAL_NUMBER = "G2"
        POLICY_FORM_IDENTIFYING_NUMBER = "0K"

    reference_identification_qualifier: ReferenceIdentificationQualifier


class Loop2110AmtSegment(AmtSegment):
    """
    Service Line Payment Information Amount
    """

    class AmountQualifierCode(str, Enum):
        """
        Code values for AMT01
        """

        ALLOWED_ACTUAL = "B6"
        DEDUCTION_AMOUNT = "KH"
        TAX = "T"
        TOTAL_CLAIM_BEFORE_TAXES = "T2"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_1 = "ZK"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_2 = "ZL"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_3 = "ZM"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_4 = "ZN"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_5 = "ZO"


class Loop2110QtySegment(QtySegment):
    """
    Service Line Payment Information Quantity
    """

    class QuantityQualifier(str, Enum):
        """
        Code values for QTY01
        """

        FEDERAL_PAYMENT_MANDATE_CATEGORY_1 = "ZK"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_2 = "ZL"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_3 = "ZM"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_4 = "ZN"
        FEDERAL_PAYMENT_MANDATE_CATEGORY_5 = "ZO"

    quantity_qualifier: QuantityQualifier


class Loop2110LqSegment(LqSegment):
    """
    Service Line Payment Information Health Care Remark Codes
    """

    class CodeListQualifierCodes(str, Enum):
        """
        Code values for LQ01
        """

        CLAIM_PAYMENT_REMARK_CODES = "HE"
        NATIONAL_COUNCIL_PRESCRIPTION_DRUG_REJECT_PAYMENT_CODES = "RX"

    code_list_qualifier_code: CodeListQualifierCodes



"""
test_support.py

Tests support functions.
"""
import pytest
import datetime
from x12.support import (
    is_x12_data,
    is_x12_file,
    parse_x12_date,
    parse_interchange_date,
    count_segments,
)
from x12.io import X12ModelReader


@pytest.mark.parametrize(
    "test_input, is_fixture, expected",
    [
        ("simple_270_with_new_lines", True, True),
        ("simple_270_one_line", True, True),
        ("foo", False, False),
        ("", False, False),
        (None, False, False),
    ],
)
def test_is_x12_data(request, test_input: str, is_fixture: bool, expected: bool):
    """
    Tests is_x12_data with fixtures and literal values.
    :param request: The pytest request fixture. Used to lookup fixture values by name.
    :param test_input: The fixture name or the literal value.
    :param is_fixture: Indicates if test_input is a fixture name.
    :param expected: The expected test value
    """

    input_value = request.getfixturevalue(test_input) if is_fixture else test_input
    assert is_x12_data(input_value) is expected


@pytest.mark.parametrize(
    "test_input", ["simple_270_with_new_lines", "simple_270_one_line"]
)
def test_is_x12_file_true(request, tmpdir, test_input: str):
    """
    Tests is_x12_file where the expected result is True.
    :param request: The pytest request fixture. Used to lookup fixture values by name.
    :param tmpdir: The pytest tmpdir fixture. Used to create tmp directory and files.
    :param test_input: The fixture name.
    """

    input_value = request.getfixturevalue(test_input)
    f = tmpdir.mkdir("x12-support").join("test.x12")
    f.write(input_value)

    assert is_x12_file(f)


def test_is_x12_file_false():
    assert is_x12_file("/home/not-a-real/file.txt") is False
    assert is_x12_file("") is False
    assert is_x12_file(None) is False


def test_parser_interchange_date():
    assert parse_interchange_date("131031") == datetime.date(2013, 10, 31)


def test_parse_x12_date():
    assert parse_x12_date("20120501") == datetime.date(2012, 5, 1)
    assert parse_x12_date("201205011010") == datetime.datetime(2012, 5, 1, 10, 10)
    assert parse_x12_date("") is None
    assert parse_x12_date(None) is None


def test_count_segments(simple_270_with_new_lines):
    with X12ModelReader(simple_270_with_new_lines) as r:
        model = [m for m in r.models()][0]
        model_data = model.dict()
        segment_count: int = count_segments(model_data)
        assert segment_count == 17
